/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 2002 The Apache Software Foundation.  All rights 
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:  
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "WSIF" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written 
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 2001, 2002, International
 * Business Machines, Inc., http://www.apache.org.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

package org.apache.wsif.providers.soap.apachesoap;

import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import javax.wsdl.Binding;
import javax.wsdl.Definition;
import javax.wsdl.Port;
import javax.wsdl.Service;
import javax.wsdl.extensions.soap.SOAPBinding;

import org.apache.wsif.WSIFException;
import org.apache.wsif.WSIFPort;
import org.apache.wsif.base.WSIFServiceImpl;
import org.apache.wsif.logging.Trc;
import org.apache.wsif.providers.WSIFDynamicTypeMap;
import org.apache.wsif.spi.WSIFProvider;

/**
 * ApacheSOAP specific provider of dynamic WSDL invocations.
 *
 * @author Aleksander Slominski
 * @author Owen Burroughs <owenb@apache.org>
 * @author Ant Elder <antelder@apache.org>
 * @author Jeremy Hughes <hughesj@apache.org>
 * @author Mark Whitlock <whitlock@apache.org>

 */
public class WSIFDynamicProvider_ApacheSOAP implements WSIFProvider {
    protected String partSerializerName = null;

    private static final String soap = "http://schemas.xmlsoap.org/wsdl/soap/";
    private static final String jms = "http://schemas.xmlsoap.org/wsdl/jms/";
    private static String[] bindings = new String[0];
    private static String[] addresses = new String[0];
    private static boolean setUpBindings = false;
    private static boolean setUpAddresses = false;

    public WSIFDynamicProvider_ApacheSOAP() {
        Trc.entry(this);
        if (!setUpBindings) {
            setUpBindingNamespaceURIs();
        }
        if (!setUpAddresses) {
            setUpAddressNamespaceURIs();
        }
        WSIFServiceImpl.addExtensionRegistry(
            new org.apache.wsif.wsdl.extensions.jms.JMSExtensionRegistry());
        Trc.exit();
    }

    /**
     * Check if WSDL port has SOAP binding and if successful try
     * to create ApacheSOAP port instance.
     */
    public WSIFPort createDynamicWSIFPort(
        Definition def,
        Service service,
        Port port,
        WSIFDynamicTypeMap typeMap)
        throws WSIFException {
        Trc.entry(this, def, service, port, typeMap);

        // check that Port binding has SOAP binding extensibility element
        Binding binding = port.getBinding();
        List exs = binding.getExtensibilityElements();
        for (Iterator i = exs.iterator(); i.hasNext();) {
            Object o = i.next();
            if (o instanceof SOAPBinding) {
                // if so try to create ApacheSOAP dynamic port instance
                WSIFPort wp =
                    new WSIFPort_ApacheSOAP(
                        def,
                        service,
                        port,
                        typeMap,
                        partSerializerName);
                Trc.exit(wp);
                return wp;
            }
        }

        // otherwise return null (so other providers can be checked)
        Trc.exit(null);
        return null;
    }

    /**
     * Gets the partSerializerName.
     * @return Returns a String
     */
    public String getPartSerializerName() {
        Trc.entry(this);
        Trc.exit(partSerializerName);
        return partSerializerName;
    }

    /**
     * Sets the partSerializerName.
     * @param partSerializerName The partSerializerName to set
     */
    public void setPartSerializerName(String partSerializerName) {
        Trc.entry(this, partSerializerName);
        this.partSerializerName = partSerializerName;
        Trc.exit();
    }

    /**
     * Returns the WSDL namespace URIs of any bindings this provider supports.
     * @return an array of all binding namespaces supported by this provider
     */
    public String[] getBindingNamespaceURIs() {
        Trc.entry(this);
        Trc.exit(bindings);
        return bindings;
    }

    /**
     * Returns the WSDL namespace URIs of any port addresses this provider supports.
     * @return an array of all address namespaces supported by this provider
     */
    public String[] getAddressNamespaceURIs() {
        Trc.entry(this);
        Trc.exit(addresses);
        return addresses;
    }

    private void setUpBindingNamespaceURIs() {
        // check if the soap classes are available, if not then we cannot
        // support soap bindings!        
        Class cls =
            (Class) AccessController.doPrivileged(new PrivilegedAction() {
            public Object run() {
                try {
                    return Class.forName(
                        "org.apache.soap.Constants",
                        true,
                        Thread.currentThread().getContextClassLoader());
                } catch (Throwable ignored) {
                   	Trc.ignoredException(ignored);
                }
                return null;
            }
        });
        if (cls != null) {
            bindings = new String[] { soap };
        }
        setUpBindings = true;
    }

    private void setUpAddressNamespaceURIs() {
        Vector v = new Vector();
        // check if the jms classes are available, if not then we cannot
        // support jms addresses!        
        Class cls =
            (Class) AccessController.doPrivileged(new PrivilegedAction() {
            public Object run() {
                try {
                    return Class.forName(
                        "javax.jms.Queue",
                        true,
                        Thread.currentThread().getContextClassLoader());
                } catch (Throwable ignored) {
                  	Trc.ignoredException(ignored);
                }
                return null;
            }
        });
        if (cls != null) {
        	v.add(jms);
        }
                        
        // check if the soap classes are available, if not then we cannot
        // support soap addresses!         
        Class cls2 =
            (Class) AccessController.doPrivileged(new PrivilegedAction() {
            public Object run() {
                try {
                    return Class.forName(
                        "org.apache.soap.Constants",
                        true,
                        Thread.currentThread().getContextClassLoader());
                } catch (Throwable ignored) {
                 	Trc.ignoredException(ignored);
                }
                return null;
            }
        });
        if (cls2 != null) {
        	v.add(soap);
        }         
        addresses = null;
        addresses = new String[v.size()];
        for (int i = 0; i < v.size(); i++) {
            addresses[i] = (String) v.elementAt(i);
        }
        setUpAddresses = true;
    }
}