/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 2002 The Apache Software Foundation.  All rights 
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:  
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "WSIF" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written 
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 2001, 2002, International
 * Business Machines, Inc., http://www.apache.org.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

package mime;

import java.awt.Image;
import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.Arrays;

import javax.activation.DataHandler;
import javax.activation.FileDataSource;
import javax.mail.internet.MimeMultipart;
import javax.swing.ImageIcon;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;
import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import org.apache.wsif.WSIFException;
import org.apache.wsif.WSIFMessage;
import org.apache.wsif.WSIFOperation;
import org.apache.wsif.WSIFPort;
import org.apache.wsif.WSIFService;
import org.apache.wsif.WSIFServiceFactory;
import org.apache.wsif.util.WSIFPluggableProviders;
import util.TestUtilities;

/**
 * Junit test for Mime attachments
 * 
 * There should be tests here for...
 * DataHandler, DataSource and other datatypes (not sure of the list)
 * Sync and async
 * rpc-style and doc-style
 * Other mime types
 * Axis and Soap 2.3
 * Soap Http and Jms
 * Does this function make sense for NativeJms, Java, EJB, etc?
 * Sending attachments and receiving them
 * Send or receive a null attachment
 * Sending or receiving an object that extends DataHandler, DataSource, etc
 * Multiple mime parts, multiple soap bodies, or no soap body at all.
 * Is the Mime type that is specified in the WSDL passed into Axis?
 * 
 * @author Mark Whitlock
 */
public class MimeTest extends TestCase {
    private final static String wsdlPath =
        TestUtilities.getWsdlPath("java\\test\\mime");
    private final static String imageLocation = wsdlPath + "axis.jpg";
    private final static String flatfileLocation = wsdlPath + "test.txt";
    private final static String flatfileLocation2 = wsdlPath + "test2.txt";
    private final static String portName =
        TestUtilities.getSoapServer().toUpperCase() + "Port";

    private final static String SEND_DH = "SEND-DH";
    private final static String RECEIVE_DH = "RECEIVE-DH";
    private final static String SEND_PLAINTEXT = "SEND-PLAINTEXT";
    private final static String RECEIVE_PLAINTEXT = "RECEIVE-PLAINTEXT";
    private final static String BOUNCE_IMAGE = "BOUNCE-IMAGE";
    private final static String BOUNCE_IMAGE2 = "BOUNCE-IMAGE2";
    private final static String BOUNCE_IMAGE3 = "BOUNCE-IMAGE3";
    private final static String BOUNCE_IMAGE4_DEFAULT = "BOUNCE-IMAGE4-DEFAULT";
    private final static String BOUNCE_IMAGE4_FALSE = "BOUNCE-IMAGE4-FALSE";
    private final static String BOUNCE_IMAGE4_NULL = "BOUNCE-IMAGE4-NULL";
    private final static String SEND_SOURCE = "SEND-SOURCE";
    private final static String RECEIVE_SOURCE = "RECEIVE-SOURCE";
    private final static String SEND_MIMEMULTIPART = "SEND-MIMEMULTIPART";
    private final static String RECEIVE_MIMEMULTIPART = "RECEIVE-MIMEMULTIPART";
    private final static String OR_MULTIPARTS1 = "OR-MULTIPARTS1";
    private final static String OR_MULTIPARTS2 = "OR-MULTIPARTS2";
    private final static String AND_MULTIPARTS = "AND-MULTIPARTS";
    private final static String MULTI_OUT_PARTS = "MULTI-OUT-PARTS";
    private final static String MULTI_INOUT_PARTS = "MULTI-INOUT-PARTS";
    private final static String NO_CONTENT = "NO-CONTENT";
    private final static String TYPE_STAR = "TYPE-STAR";
    private final static String SOAP_BODY_PARTS1 = "SOAP_BODY_PARTS1";
    private final static String SOAP_BODY_PARTS2 = "SOAP_BODY_PARTS2";
    private final static String SOAP_BODY_PARTS3 = "SOAP_BODY_PARTS3";
    private final static String SOAP_BODY_PARTS4 = "SOAP_BODY_PARTS4";
    private final static String ARRAY_OF_BINARY = "ARRAY-OF-BINARY";
    private final static String OPTIONAL_SOAP_BODY = "OPTIONAL_SOAP_BODY";
    private final static String MIX_MIME_PARTS = "MIX_MIME_PARTS";
    private final static String MAP_TYPE = "MAP-TYPE";
    private final static String BAD_NO_PART = "BAD-NO-PART";
    private final static String BAD_PART = "BAD-PART";
    private final static String BAD_NESTED = "BAD-NESTED";
    private final static String BAD_MIX_SOAP_MIME = "BAD-MIX-SOAP-MIME";
    private final static String BAD_MULTIPLE_SOAP_BODIES = "BAD-MULTIPLE-SOAP-BODIES";
    private final static String BAD_SOAP_BODY_TYPE = "BAD-SOAP-BODY-TYPE";

    /** 
     * Milliseconds to sleep while the user looks at the displayed images.
     */
    private int SLEEPY_TIME = 0;
    
    private final static String[] rhyme =
        { "The owl and the pussy cat went to sea in a beautiful pea-green boat,",
          "They took some honey and plenty of money wrapped up in a five pound note" };

    private final boolean SYNC = true;
    private final boolean ASYNC = false;
    
    public MimeTest(String name) {
        super(name);

        try {
            SLEEPY_TIME =
                Integer.parseInt(
                    TestUtilities.getWsifProperty("wsif.displaytime"));
        } finally {
            if (SLEEPY_TIME <= 0)
                SLEEPY_TIME = 2000;
        }
    }

    public static void main(String[] args) {
	   // TestUtilities.startListeners(TestUtilities.ADDRESSBOOK_LISTENER);	
	   junit.textui.TestRunner.run (suite());
	   // TestUtilities.stopListeners();	
    }

    public static Test suite() {
        return new TestSuite(MimeTest.class);
    }

    public void setUp() {
        TestUtilities.setUpExtensionsAndProviders();
    }

    public void testSendHandler() {
        doit(SEND_DH, "Mime.wsdl");
    }

    public void testReceiveHandler() {
        doit(RECEIVE_DH, "Mime.wsdl");
    }

    /*
     * Attachments that are Strings, Images, are not supported at present.
     * Attachments can only be DataHandlers.
     * 
     * public void testSendPlainText() {
     *     doit(SEND_PLAINTEXT, "Mime.wsdl");
     * }
     *
     * public void testReceivePlainText() {
     *    doit(RECEIVE_PLAINTEXT, "Mime.wsdl");
     * }
     * 
     * public void testBounceImage() {
     *     doit(BOUNCE_IMAGE, "Mime.wsdl");
     * }
     */

    public void testBounceImage2() {
        doit(BOUNCE_IMAGE2, "Mime.wsdl");
    }

    public void testBounceImage3() {
        doit(BOUNCE_IMAGE3, "Mime.wsdl");
    }

    public void testBounceImage4Default() {
        doit(BOUNCE_IMAGE4_DEFAULT, "Mime.wsdl");
    }
    
    public void testBounceImage4False() {
        doit(BOUNCE_IMAGE4_FALSE, "Mime.wsdl");
    }

    /*
     * This next test fails because axis will not allow null 
     * parameters to be passed in.
     * 
     * public void testBounceImage4Null() {
     *     doit(BOUNCE_IMAGE4_NULL, "Mime.wsdl");
     * }
     */

    /*
     * bounceImage5 doesn't work because the soap:body (which contains a 
     *    boolean) is within the mime:part. bounceImage4 works because the
     *    soap:body is a direct child of the binding input. When I try to 
     *    run bounceImage5, WSDL4J says the soap:body doesn't contain any
     *    parts. But having the soap:body inside the mime:part is valid WSDL
     *    (according to the wsdl4j PopulatedExtensionRegistry). So what does
     *    it mean? 
     */

    /*
     * public void testSendSource() {
     *     doit(SEND_SOURCE, "Mime.wsdl");
     * }
     *
     * public void testReceiveSource() {
     *     doit(RECEIVE_SOURCE, "Mime.wsdl");
     * }
     *
     * public void testSendMimeMultipart() {
     *     doit(SEND_MIMEMULTIPART, "Mime.wsdl");
     * }
     *
     * public void testReceiveMimeMultipart() {
     *     doit(RECEIVE_MIMEMULTIPART, "Mime.wsdl");
     * }
     */

    /*
     * Sending MIME messages using Axis over Jms should be supported 
     * by WSIF, but currently this does not work. I think the problem
     * is that WSIFJmsSender does not set up the necessary headers needed
     * for MIME. I don't know whether these headers can go inline in the
     * message or whether they will have to go as JMS properties which 
     * will then get converted into HTTP headers by the JMS2HTTPBridge.
     * For more information about the headers please see 
     * org.apache.axis.transport.http.HTTPSender. For the moment I have 
     * commented out this Jms test (and omitted other JMS tests like 
     * the async tests) because it fails.
     *
     * public void testJmsSend() {
     *   doit("SOAPJMSPort", SEND, "Mime.wsdl");
     * }
     */

    public void testOrMultiParts1() {
        doit(OR_MULTIPARTS1, "Mime.wsdl");
    }

    public void testOrMultiParts2() {
        doit(OR_MULTIPARTS2, "Mime.wsdl");
    }

    public void testAndMultiParts() {
        doit(AND_MULTIPARTS, "Mime.wsdl");
    }

    /*
     *  public void testMultiOutParts() {
     *     doit(MULTI_OUT_PARTS, "Mime.wsdl");
     * }
     *
     * public void testMultiInoutParts() {
     *     doit(MULTI_INOUT_PARTS, "Mime.wsdl");
     * }
     */

    public void testNoContent() {
        doit(NO_CONTENT, "Mime.wsdl");
    }

    public void testTypeStar() {
        doit(TYPE_STAR, "Mime.wsdl");
    }

    public void testSoapBodyParts1() {
        doit(SOAP_BODY_PARTS1, "Mime.wsdl");
    }

    public void testSoapBodyParts2() {
        doit(SOAP_BODY_PARTS2, "Mime.wsdl");
    }

    public void testSoapBodyParts3() {
        doit(SOAP_BODY_PARTS3, "Mime.wsdl");
    }

    public void testSoapBodyParts4() {
        doit(SOAP_BODY_PARTS4, "Mime.wsdl");
    }

    public void testArrayOfBinary() {
        doit(ARRAY_OF_BINARY, "Mime.wsdl");
    }

    public void testMapType() {
        doit(MAP_TYPE, "Mime.wsdl");
    }

    public void testOptionalSoapBody() {
        doit(OPTIONAL_SOAP_BODY, "Mime.wsdl");
    }

    public void testMixMimeParts() {
        doit(MIX_MIME_PARTS, "Mime.wsdl");
    }

    /* ************ ERRORS **************** */

    public void testBadNoPart() {
        doit(BAD_NO_PART, "MimeBadNoPart.wsdl");
    }

    public void testBadPart() {
        doit(BAD_PART, "MimeBadPart.wsdl");
    }

    public void testBadNested() {
        doit(BAD_NESTED, "MimeBadNested.wsdl");
    }

    public void testBadMixSoapMime() {
        doit(BAD_MIX_SOAP_MIME, "MimeBadMixSoapMime.wsdl");
    }

    public void testBadMultipleSoapBodies() {
        doit(BAD_MULTIPLE_SOAP_BODIES, "MimeBadMultipleSoapBodies.wsdl");
    }

    public void testBadSoapBodyType() {
        doit(BAD_SOAP_BODY_TYPE, "MimeBadSoapBodyType.wsdl");
    }

    /**
     * doit should probably do the mapTypes() but unfortunately plain text
     * means that Strings are serialized with the JAFDataHandlerSerializer
     * and I can't find a way to differentiate between strings that should
     * be treated as a mime part and those that should not. Basically if I
     * mapTypes(plaintext,String) all Strings get mapped to a DataHandler,
     * and soap on the server looks for a web service that takes a 
     * DataHandler, not a String. So be careful where mapTypes(plaintext,String)
     * is done.
     */
    private void doit(String cmd, String wsdl) {
        if (!TestUtilities.areWeTesting("mime"))
            return;
            
        WSIFPluggableProviders.overrideDefaultProvider(
            "http://schemas.xmlsoap.org/wsdl/soap/",
            new org.apache.wsif.providers.soap.apacheaxis
                .WSIFDynamicProvider_ApacheAxis());

        try {
            WSIFServiceFactory factory = WSIFServiceFactory.newInstance();
            WSIFService service =
                factory.getService(
                    wsdlPath + wsdl,
                    null,
                    null,
                    "http://mime/",
                    "Mime");

            // Many of the error tests are badly behaved and would
            // throw an exception here.
            Mime stub = null;
            if (!cmd.startsWith("BAD"))
                stub = (Mime) service.getStub(portName, Mime.class);

            if (cmd.equals(SEND_DH))
                send_dh(service,stub);
            else if (cmd.equals(RECEIVE_DH))
                receive_dh(service,stub);
            else if (cmd.equals(SEND_PLAINTEXT))
                send_plaintext(service,stub);
            else if (cmd.equals(RECEIVE_PLAINTEXT))
                receive_plaintext(service,stub);
            else if (cmd.equals(BOUNCE_IMAGE))
                bounce_image(service,stub);
            else if (cmd.equals(BOUNCE_IMAGE2))
                bounce_image2(service,stub,portName);
            else if (cmd.equals(BOUNCE_IMAGE3))
                bounce_image3(service,stub,portName);
            else if (cmd.equals(BOUNCE_IMAGE4_DEFAULT))
                bounce_image4_default(service,stub,portName);
            else if (cmd.equals(BOUNCE_IMAGE4_FALSE))
                bounce_image4_false(service,stub,portName);
            else if (cmd.equals(BOUNCE_IMAGE4_NULL))
                bounce_image4_null(service,stub,portName);
            else if (cmd.equals(SEND_SOURCE))
                send_source(service,stub);
            else if (cmd.equals(RECEIVE_SOURCE))
                receive_source(service,stub);
            else if (cmd.equals(SEND_MIMEMULTIPART))
                send_mimemultipart(service,stub);
            else if (cmd.equals(RECEIVE_MIMEMULTIPART))
                receive_mimemultipart(service,stub);
            else if (cmd.equals(OR_MULTIPARTS1))
                or_multiparts1(service,stub);
            else if (cmd.equals(OR_MULTIPARTS2))
                or_multiparts2(service,stub);
            else if (cmd.equals(AND_MULTIPARTS))
                and_multiparts(service,stub);
            else if (cmd.equals(MULTI_OUT_PARTS))
                multi_out_parts(service,stub);
            else if (cmd.equals(MULTI_INOUT_PARTS))
                multi_inout_parts(service,stub);
            else if (cmd.equals(NO_CONTENT))
                no_content(service,stub);
            else if (cmd.equals(TYPE_STAR))
                type_star(service,stub);
            else if (cmd.equals(SOAP_BODY_PARTS1))
                soap_body_parts1(service,stub);
            else if (cmd.equals(SOAP_BODY_PARTS2))
                soap_body_parts2(service,stub);
            else if (cmd.equals(SOAP_BODY_PARTS3))
                soap_body_parts3(service,stub);
            else if (cmd.equals(SOAP_BODY_PARTS4))
                soap_body_parts4(service,stub);
            else if (cmd.equals(ARRAY_OF_BINARY))
                array_of_binary(service,stub);
            else if (cmd.equals(MAP_TYPE))
                map_type(service,stub);
            else if (cmd.equals(OPTIONAL_SOAP_BODY))
                optional_soap_body(service,stub);
            else if (cmd.equals(MIX_MIME_PARTS))
                mix_mime_parts(service,stub);
            else if (cmd.equals(BAD_NO_PART))
                bad_no_part(service);
            else if (cmd.equals(BAD_PART))
                bad_part(service);
            else if (cmd.equals(BAD_NESTED))
                bad_nested(service);
            else if (cmd.equals(BAD_MIX_SOAP_MIME))
                bad_mix_soap_mime(service);
            else if (cmd.equals(BAD_MULTIPLE_SOAP_BODIES))
                bad_multiple_soap_bodies(service);
            else if (cmd.equals(BAD_SOAP_BODY_TYPE))
                bad_soap_body_type(service);
            else
                assertTrue(false);

        } catch (Exception e) {
            System.err.println(
                "MimeTest(" + portName + ") caught exception " + e);
            e.printStackTrace();
            assertTrue(false);
        } finally {
            WSIFPluggableProviders.overrideDefaultProvider(
                "http://schemas.xmlsoap.org/wsdl/soap/",
                null);
                
            WSIFFrame.close();
        }

    }

    private void send_dh(WSIFService service, Mime stub) throws Exception {
        DataHandler dh = new DataHandler(new FileDataSource(flatfileLocation));
        String buff = stub.dataHandlerToString(dh);
        assertTrue(compareFiles(flatfileLocation, buff));
    }

    private void receive_dh(WSIFService service, Mime stub) throws Exception {
        DataHandler dh = stub.stringToDataHandler(rhyme[0]);
        assertTrue(compareFiles(dh, rhyme[0]));
    }

    private void send_plaintext(WSIFService service, Mime stub)
        throws Exception {
        service.mapType(
            new QName("http://mime/", "plaintext"),
            String.class);
        String buff = stub.plainTextToString(rhyme[0]);
        assertTrue(rhyme[0].equals(buff));
    }

    private void receive_plaintext(WSIFService service, Mime stub)
        throws Exception {
        service.mapType(
            new QName("http://mime/", "plaintext"),
            String.class);
        String buff = stub.stringToPlainText(rhyme[1]);
        assertTrue(rhyme[1].equals(buff));
    }

    private void bounce_image(WSIFService service, Mime stub) throws Exception {
        service.mapType(
            new QName("http://mime/", "image"),
            Image.class);

        // This blocks until the image is loaded.
        Image im1 = new ImageIcon(imageLocation).getImage();
        WSIFFrame.display(im1,"Original image");

        // Could use Image image2 = Toolkit.getDefaultToolkit().getImage("image.gif");
        // which loads the image in the background.

        Image im2 = stub.bounceImage(im1);
        WSIFFrame.display(im2,"Bounced image");

    	System.out.println("Sleeping");
    	Thread.sleep(SLEEPY_TIME);
    	System.out.println("Woken up");
    }

    /**
     * This test does not do any mapTypes() to test that WSIF will 
     * automatically register mime parts as a DataHandler. 
     */
    private void bounce_image2(WSIFService service, Mime stub, String portName)
        throws Exception {
        DataHandler dh1 = new DataHandler(new FileDataSource(imageLocation));

        WSIFPort port = service.getPort(portName);
        WSIFOperation op = port.createOperation("bounceImage2");
        WSIFMessage in = op.createInputMessage();
        WSIFMessage out = op.createOutputMessage();
        WSIFMessage fault = op.createFaultMessage();
        in.setObjectPart("file",dh1);

        boolean success = op.executeRequestResponseOperation(in,out,fault);
        assertTrue(success);

        DataHandler dh2 = (DataHandler)(out.getObjectPart("file2"));

//        This commented out code displays the image, so proving that the 
//        bounced image is correct.
//
//        InputStream is = dh2.getInputStream();
//        byte[] bBuff = new byte[is.available()];
//        is.read(bBuff);
//        Image im = new ImageIcon(bBuff).getImage();
//        WSIFFrame.display(im,"Image");
        
        assertTrue(compareFiles(dh1,dh2));
    }

    private void bounce_image3(WSIFService service, Mime stub, String portName)
        throws Exception {
        DataHandler dh1 = new DataHandler(new FileDataSource(imageLocation));
        DataHandler dh2 = stub.bounceImage2(dh1);
        assertTrue(compareFiles(dh1,dh2));
    }

    private void bounce_image4_default(WSIFService service, Mime stub, String portName)
        throws Exception {
        DataHandler dh1 = new DataHandler(new FileDataSource(imageLocation));
        DataHandler dh2 = stub.bounceImage4(true, dh1); 
        assertTrue(compareFiles(dh1,dh2));
    }

    private void bounce_image4_false(WSIFService service, Mime stub, String portName)
        throws Exception {
        DataHandler dh1 = new DataHandler(new FileDataSource(imageLocation));
        DataHandler dh2 = stub.bounceImage4(false, dh1);
        assertTrue(dh2==null);
    }

    private void bounce_image4_null(WSIFService service, Mime stub, String portName)
        throws Exception {
        DataHandler dh2 = stub.bounceImage4(true, null);
        assertTrue(dh2==null); 
    }

    private void send_source(WSIFService service, Mime stub) throws Exception {
            service.mapType(
                new QName("http://mime/", "source"),
                Source.class);
//        String buff = stub.plainTextToString(rhyme[0]);
//        assertTrue(rhyme[0].equals(buff));
        assertTrue(false);
    }

    private void receive_source(WSIFService service, Mime stub) throws Exception {
            service.mapType(
                new QName("http://mime/", "source"),
                Source.class);
//        String buff = stub.stringToPlainText(rhyme[1]);
//        assertTrue(rhyme[1].equals(buff));
        assertTrue(false);
    }

    private void send_mimemultipart(WSIFService service, Mime stub) throws Exception {
            service.mapType(
                new QName("http://mime/", "mimemultipart"),
                MimeMultipart.class);
//        String buff = stub.plainTextToString(rhyme[0]);
//        assertTrue(rhyme[0].equals(buff));
        assertTrue(false);
    }

    private void receive_mimemultipart(WSIFService service, Mime stub) throws Exception {
            service.mapType(
                new QName("http://mime/", "mimemultipart"),
                MimeMultipart.class);
//        String buff = stub.stringToPlainText(rhyme[1]);
//        assertTrue(rhyme[1].equals(buff));
        assertTrue(false);
    }

    private void or_multiparts1(WSIFService service, Mime stub) throws Exception {
        DataHandler dh = new DataHandler(new FileDataSource(flatfileLocation));
        String buff = stub.orMultiMimeParts(dh);
        assertTrue("text/plain".equals(buff));
    }

    private void or_multiparts2(WSIFService service, Mime stub) throws Exception {
        DataHandler dh = new DataHandler(new FileDataSource(imageLocation));
        String buff = stub.orMultiMimeParts(dh);
        assertTrue("image/jpeg".equals(buff));
    }

    private void and_multiparts(WSIFService service, Mime stub) throws Exception {
        DataHandler dh1 = new DataHandler(new FileDataSource(flatfileLocation));
        DataHandler dh2 = new DataHandler(new FileDataSource(flatfileLocation2));
        String buff = stub.andMultiMimeParts(dh1,dh2);
        assertTrue(concat(flatfileLocation,flatfileLocation2).equals(buff));
    }

    private void multi_out_parts(WSIFService service, Mime stub) throws Exception {
        assertTrue(false);
    }

    private void multi_inout_parts(WSIFService service, Mime stub) throws Exception {
        assertTrue(false);
    }

    private void no_content(WSIFService service, Mime stub) throws Exception {
        assertTrue(rhyme[0].equals(stub.noContent(rhyme[0])));
    }

    private void type_star(WSIFService service, Mime stub) throws Exception {
        DataHandler dh = new DataHandler(new FileDataSource(flatfileLocation));
        String buff = stub.dataHandlerToString(dh);
        assertTrue(compareFiles(flatfileLocation, buff));
    }

    /**
     * These next tests have to be run stubless because they test the parts="..."
     * in the soap:body and the WSIFClientProxy doesn't look in the binding
     * and so would not be able to find this signature in the portType.
     */
    private void soap_body_parts1(WSIFService service, Mime stub)
        throws Exception {
        DataHandler dh = new DataHandler(new FileDataSource(flatfileLocation));
        WSIFPort port = service.getPort(portName);
        WSIFOperation op = port.createOperation("soapBodyParts1");
        WSIFMessage in = op.createInputMessage();
        WSIFMessage out = op.createOutputMessage();
        WSIFMessage fault = op.createFaultMessage();
        in.setBooleanPart("shouldBounce",true);
        in.setObjectPart("file",dh);

        boolean success = op.executeRequestResponseOperation(in,out,fault);
        assertTrue(success);

        String s = (String)(out.getObjectPart("buff"));
        assertTrue("1".equals(s));
    }

    private void soap_body_parts2(WSIFService service, Mime stub)
        throws Exception {
        DataHandler dh = new DataHandler(new FileDataSource(flatfileLocation));
        WSIFPort port = service.getPort(portName);
        WSIFOperation op = port.createOperation("soapBodyParts2");
        WSIFMessage in = op.createInputMessage();
        WSIFMessage out = op.createOutputMessage();
        WSIFMessage fault = op.createFaultMessage();
        in.setObjectPart("file",dh);

        boolean success = op.executeRequestResponseOperation(in,out,fault);
        assertTrue(success);

        String s = (String)(out.getObjectPart("buff"));
        assertTrue("2".equals(s));
    }

    private void soap_body_parts3(WSIFService service, Mime stub)
        throws Exception {
        WSIFPort port = service.getPort(portName);
        WSIFOperation op = port.createOperation("soapBodyParts3");
        WSIFMessage in = op.createInputMessage();
        WSIFMessage out = op.createOutputMessage();
        WSIFMessage fault = op.createFaultMessage();
        in.setBooleanPart("shouldBounce",true);

        boolean success = op.executeRequestResponseOperation(in,out,fault);
        assertTrue(success);

        String s = (String)(out.getObjectPart("buff"));
        assertTrue("3".equals(s));
    }

    private void soap_body_parts4(WSIFService service, Mime stub)
        throws Exception {
        WSIFPort port = service.getPort(portName);
        WSIFOperation op = port.createOperation("soapBodyParts4");
        WSIFMessage in = op.createInputMessage();
        WSIFMessage out = op.createOutputMessage();
        WSIFMessage fault = op.createFaultMessage();

        boolean success = op.executeRequestResponseOperation(in,out,fault);
        assertTrue(success);

        String s = (String)(out.getObjectPart("buff"));
        assertTrue("4".equals(s));
    }

    private void array_of_binary(WSIFService service, Mime stub) throws Exception {
        DataHandler dh = new DataHandler(new FileDataSource(flatfileLocation));
        String buff = stub.arrayOfBinary(dh);
        assertTrue(compareFiles(flatfileLocation, buff));
    }

    private void map_type(WSIFService service, Mime stub) throws Exception {
        service.mapType(
            new QName("http://mime/", "datahandler"),
            DataHandler.class);

        DataHandler dh = new DataHandler(new FileDataSource(flatfileLocation));
        String buff = stub.dataHandlerToString(dh);
        assertTrue(compareFiles(flatfileLocation, buff));
    }

    private void optional_soap_body(WSIFService service, Mime stub)
        throws Exception {
        DataHandler dh1 = new DataHandler(new FileDataSource(imageLocation));
        DataHandler dh2 = stub.optionalSoapBody(dh1);
        assertTrue(compareFiles(dh1,dh2));
    }

    private void mix_mime_parts(WSIFService service, Mime stub)
        throws Exception
    {
        DataHandler dh1 = new DataHandler(new FileDataSource(flatfileLocation));
        DataHandler dh2 =
            new DataHandler(new FileDataSource(flatfileLocation2));
        String position1 = "Position One";
        String position2 = "Position Two";
        String position3 = "Position Three";

        WSIFPort port = service.getPort(portName);
        WSIFOperation op = port.createOperation("mixMimeParts");
        WSIFMessage in = op.createInputMessage();
        WSIFMessage out = op.createOutputMessage();
        WSIFMessage fault = op.createFaultMessage();
        in.setObjectPart("position1", position1);
        in.setObjectPart("file1", dh1);
        in.setObjectPart("position2", position2);
        in.setObjectPart("file2", dh2);
        in.setObjectPart("position3", position3);

        boolean success = op.executeRequestResponseOperation(in, out, fault);
        assertTrue(success);

        String s = (String) (out.getObjectPart("buff"));
        String expected =
            position1 + readFile(dh1) + position2 + readFile(dh2) + position3;
        assertTrue(expected.equals(s));
    }

    /* *********************** ERRORS ********************************** */
    
    private void bad_no_part(WSIFService service) throws Exception {
        DataHandler dh = new DataHandler(new FileDataSource(flatfileLocation));
        boolean exceptionCaught = false;
        try {
            Mime stub = (Mime) service.getStub(portName, Mime.class);
            String buff = stub.badNoPart(dh);
        } catch (WSIFException we) {
            exceptionCaught = true;
            System.out.println("Expected exception="+we);
            we.printStackTrace();
        }
        assertTrue(exceptionCaught);
    }

    private void bad_part(WSIFService service) throws Exception {
        DataHandler dh = new DataHandler(new FileDataSource(flatfileLocation));
        boolean exceptionCaught = false;
        try {
            Mime stub = (Mime) service.getStub(portName, Mime.class);
            String buff = stub.badPart(dh);
        } catch (WSIFException we) {
            exceptionCaught = true;
            System.out.println("Expected exception="+we);
            we.printStackTrace();
        }
        assertTrue(exceptionCaught);
    }

    private void bad_nested(WSIFService service) throws Exception {
        DataHandler dh = new DataHandler(new FileDataSource(flatfileLocation));
        boolean exceptionCaught = false;
        try {
            Mime stub = (Mime) service.getStub(portName, Mime.class);
            stub.badNested(dh);
        } catch (WSIFException we) {
            exceptionCaught = true;
            System.out.println("Expected exception="+we);
            we.printStackTrace();
        }
        assertTrue(exceptionCaught);
    }

    private void bad_mix_soap_mime(WSIFService service) throws Exception {
        DataHandler dh = new DataHandler(new FileDataSource(flatfileLocation));
        boolean exceptionCaught = false;
        try {
            Mime stub = (Mime) service.getStub(portName, Mime.class);
            String buff = stub.badMixSoapMime(dh);
        } catch (WSIFException we) {
            exceptionCaught = true;
            System.out.println("Expected exception="+we);
            we.printStackTrace();
        }
        assertTrue(exceptionCaught);
    }

    private void bad_multiple_soap_bodies(WSIFService service) throws Exception {
        DataHandler dh = new DataHandler(new FileDataSource(flatfileLocation));
        boolean exceptionCaught = false;
        try {
            Mime stub = (Mime) service.getStub(portName, Mime.class);
            String buff = stub.badMultipleSoapBodies(dh);
        } catch (WSIFException we) {
            exceptionCaught = true;
            System.out.println("Expected exception="+we);
            we.printStackTrace();
        }
        assertTrue(exceptionCaught);
    }

    private void bad_soap_body_type(WSIFService service) throws Exception {
        DataHandler dh = new DataHandler(new FileDataSource(flatfileLocation));
        boolean exceptionCaught = false;
        try {
            Mime stub = (Mime) service.getStub(portName, Mime.class);
            String buff = stub.badSoapBodyType(dh);
        } catch (WSIFException we) {
            exceptionCaught = true;
            System.out.println("Expected exception="+we);
            we.printStackTrace();
        }
        assertTrue(exceptionCaught);
    }

    /* ******************* UTILITIES ************************************* */
    
    private boolean compareFiles(DataHandler dh1, DataHandler dh2)
        throws FileNotFoundException, IOException {
        InputStream is1 = dh1.getInputStream();
        InputStream is2 = dh2.getInputStream();
        boolean success = false;
        try {
            success = compareFiles(is1, is2);
        } finally {
            if (null != is1)
                is1.close();
            if (null != is2)
                is2.close();
        }
        return success;
    }

    private boolean compareFiles(DataHandler dh, String buff)
        throws FileNotFoundException, IOException {
        InputStream is = dh.getInputStream();
        boolean success = false;
        try {
            success = compareFiles(is, buff);
        } finally {
            if (null != is)
                is.close();
        }
        return success;
    }

    private boolean compareFiles(String one, String buff)
        throws FileNotFoundException, IOException {
        BufferedInputStream oneStream = null;
        File f1 = new File(one);

        boolean success = false;
        try {
            oneStream =
                new BufferedInputStream(
                    new FileInputStream(one),
                    buff.length());
            success = compareFiles(oneStream, buff);
        } finally {
            if (null != oneStream)
                oneStream.close();
        }
        return success;
    }

    private boolean compareFiles(InputStream is, String buff)
        throws FileNotFoundException, IOException {
        return compareFiles(is, new ByteArrayInputStream(buff.getBytes()));
    }
    
    private boolean compareFiles(InputStream is1, InputStream is2) 
        throws FileNotFoundException, IOException {

        int avail1 = is1.available();
        int avail2 = is2.available();
        if (avail1 != avail2) return false;
        if (avail1==0) return true;

        byte[] buff1 = new byte[avail1];
        byte[] buff2 = new byte[avail2];

        Arrays.fill(buff1, (byte) 0);
        Arrays.fill(buff2, (byte) 0);

        int bread1 = -1;
        int bread2 = -1;
        bread1 = is1.read(buff1, 0, avail1);
        bread2 = is2.read(buff2, 0, avail2);
        String s1 = new String(buff1);
        String s2 = new String(buff2);
        if (!s1.equals(s2))
            return false;
        return true;
    }
    
    private String concat(String one, String two)
        throws FileNotFoundException, IOException {
        InputStream is1 = null;
        InputStream is2 = null;
        try {
            is1 = (new DataHandler(new FileDataSource(one))).getInputStream();
            is2 = (new DataHandler(new FileDataSource(two))).getInputStream();
            int avail1 = is1.available();
            int avail2 = is2.available();

            byte[] buff1 = new byte[avail1];
            byte[] buff2 = new byte[avail2];

            Arrays.fill(buff1, (byte) 0);
            Arrays.fill(buff2, (byte) 0);

            int bread1 = -1;
            int bread2 = -1;
            bread1 = is1.read(buff1, 0, avail1);
            bread2 = is2.read(buff2, 0, avail2);
            String s1 = new String(buff1);
            String s2 = new String(buff2);
            return s1 + s2;

        } finally {
            if (null != is1)
                is1.close();
            if (null != is2)
                is2.close();
        }
    }

    private String readFile(DataHandler dh) {
        try {
            InputStream is = dh.getInputStream();
            byte[] bBuff = new byte[is.available()];
            is.read(bBuff);
            String sBuff = new String(bBuff);
            return sBuff;
        } catch (IOException ioe) {
            ioe.printStackTrace();
            return "";
        }
    }
}
